//
// Copyright (c) 2002
// Ronald Kevin Burton
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. This software is provided solely as part of ".NET Common Language
// Runtime Unleashed" to help illustrate concepts and principles described in 
// that book. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// 
using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Reflection;
using System.CodeDom;
using System.CodeDom.Compiler;
using Microsoft.CSharp;

namespace MatrixMultiplication
{
	/// <summary>
	/// Summary description for MatrixMultiplyFixed.
	/// </summary>
	public class MatrixMultiplyFixed : MatrixMultiply
	{
		private IMatrixMultiply matrixMultiplyInterface;
		public MatrixMultiplyFixed(double [,] A, double [,] B) : base(A, B)
		{
			GenerateCode();
		}
		private void GenerateCode()
		{
			string filename = "mmf";
			Stream s = File.Open(filename + ".cs", FileMode.Create);
			StreamWriter t = new StreamWriter(s);

			// Generate code in C#
			CSharpCodeProvider provider = new CSharpCodeProvider();
			ICodeGenerator cg = provider.CreateGenerator(t);
			CodeGeneratorOptions op = new CodeGeneratorOptions();

			// Generate the comments at the beginning of the function
			CodeCommentStatement comment = new CodeCommentStatement("Matrix Multiplication (Fixed)");
			cg.GenerateCodeFromStatement(comment, t, op);

			// The class is named with a unique name
			string className = "__MatrixMultiplyFixed";
			CodeTypeDeclaration matrixClass = new CodeTypeDeclaration(className);
			// The class implements IPolynomial
			matrixClass.BaseTypes.Add("MatrixMultiplication.IMatrixMultiply");

			// Set up the Multiply function
			CodeMemberMethod multiplyMethod = new CodeMemberMethod();
			multiplyMethod.Name = "Multiply";
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression("double [,]", "A"));
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression("double [,]", "B"));
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression("double [,]", "C"));
			// workaround for bug below...
			multiplyMethod.ReturnType = new CodeTypeReference("public void");
			// BUG: This doesn't generate "public", it just leaves
			// the attribute off of the member...
			multiplyMethod.Attributes |= MemberAttributes.Public;

			StringBuilder str = new StringBuilder();

			str.Append("{");
			// Loop over the rows of A
			for(int i = 0; i < A.GetLength(0); i++)
			{
				// Loop over the columns of A and rows of B
				for(int j = 0; j < A.GetLength(1); j++)
				{
					// Loop over the columns of B
					for(int k = 0; k < B.GetLength(1); k++)
						C[i,k] += A[i,j] * B[j,k];
				}
				str.Append("{");
				for(int j = 0; j < B.GetLength(1); j++)
				{
					if(j < A.GetLength(0) - 1)
						str.Append(string.Format("{0},\r\n",C[i,j]));
					else
						str.Append(string.Format("{0}",C[i,j]));
				}
				if(i < A.GetLength(0) - 1)
					str.Append("},\r\n");
				else
					str.Append("}");
			}
			str.Append("}");

			CodeSnippetTypeMember variable = new CodeSnippetTypeMember(string.Format("static readonly double [,] result = new double[{0},{1}]\r\n{2};", A.GetLength(0), B.GetLength(1), str.ToString()));
			matrixClass.Members.Add(variable);

			CodeAssignStatement assignment = new CodeAssignStatement();
			assignment.Left = new CodeVariableReferenceExpression("C");
			assignment.Right = new CodeVariableReferenceExpression("result");
			multiplyMethod.Statements.Add(assignment);

			matrixClass.Members.Add(multiplyMethod);
			cg.GenerateCodeFromType(matrixClass, t, op);

			t.Close();
			s.Close();

			// Now use the CodeProvider interface
			CSharpCodeProvider cscp = new CSharpCodeProvider();
			CompilerParameters param = new CompilerParameters(null, null, true);
			param.GenerateInMemory = true;
			param.IncludeDebugInformation = false;
			param.CompilerOptions = "/optimize+";
			param.ReferencedAssemblies.Add("MatrixMultiplication.exe");
			ICodeCompiler cc = cscp.CreateCompiler();
			CompilerResults cr = cc.CompileAssemblyFromFile(param, filename + ".cs");
			System.Collections.Specialized.StringCollection output = cr.Output;
			foreach(string outputSting in output)
			{
				Console.WriteLine(outputSting);
			}

			if(cr.Errors.Count != 0)
			{
				CompilerErrorCollection es = cr.Errors;
				foreach(System.CodeDom.Compiler.CompilerError e in es)
				{
					Console.WriteLine(e.ToString());
				}
			}
			else
			{
				matrixMultiplyInterface = (IMatrixMultiply)cr.CompiledAssembly.CreateInstance(className);
			}
		}
		public override double [,] Multiply()
		{
			if (matrixMultiplyInterface == null)
				throw new ArgumentOutOfRangeException("Could not compile fixed matrix code.");
			matrixMultiplyInterface.Multiply(A, B, C);
			return C;
		}
	}
}
