//
// Copyright (c) 2002
// Ronald Kevin Burton
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. This software is provided solely as part of ".NET Common Language
// Runtime Unleashed" to help illustrate concepts and principles described in 
// that book. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// 
using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Reflection;
using System.CodeDom;
using System.CodeDom.Compiler;
using Microsoft.CSharp;

namespace MatrixMultiplication
{
	/// <summary>
	/// Summary description for MatrixMultiplyLoop.
	/// </summary>
	public class MatrixMultiplyLoop : MatrixMultiply
	{
		private IMatrixMultiply matrixMultiplyInterface;
		public MatrixMultiplyLoop(double [,] A, double [,] B) : base(A,B)
		{
			GenerateCode();
		}
		private void GenerateCode()
		{
			string filename = "mml";
			Stream s = File.Open(filename + ".cs", FileMode.Create);
			StreamWriter t = new StreamWriter(s);

			// Generate code in C#
			CSharpCodeProvider provider = new CSharpCodeProvider();
			ICodeGenerator cg = provider.CreateGenerator(t);
			CodeGeneratorOptions op = new CodeGeneratorOptions();

			// Generate the comments at the beginning of the function
			CodeCommentStatement comment = new CodeCommentStatement("Matrix Multiplication (Loop)");
			cg.GenerateCodeFromStatement(comment, t, op);

			// The class is named with a unique name
			string className = "__MatrixMultiplyLoop";
			CodeTypeDeclaration matrixClass = new CodeTypeDeclaration(className);
			// The class implements IPolynomial
			matrixClass.BaseTypes.Add("MatrixMultiplication.IMatrixMultiply");

			// Set up the Multiply function
			CodeMemberMethod multiplyHelperMethod = new CodeMemberMethod();
			multiplyHelperMethod.Name = "_Multiply";
			multiplyHelperMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(double [,]), "A"));
			multiplyHelperMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(double [,]), "B"));
			multiplyHelperMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(int), "i"));
			multiplyHelperMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(int), "j"));
			multiplyHelperMethod.ReturnType = new CodeTypeReference(typeof(double));
			{
				CodeBinaryOperatorExpression plus;
				CodeBinaryOperatorExpression lastplus;
				plus = new CodeBinaryOperatorExpression();
				plus.Operator = CodeBinaryOperatorType.Add;
				CodeMethodReturnStatement returnStatement = new CodeMethodReturnStatement(plus);

				CodeBinaryOperatorExpression multiply;
				multiply = new CodeBinaryOperatorExpression();
				multiply.Operator = CodeBinaryOperatorType.Multiply;
				multiply.Right = new CodeSnippetExpression("A[i,0]");
				multiply.Left = new CodeSnippetExpression("B[0,j]");

				plus.Left = multiply;
				lastplus = plus;
				for(int k = 1; k < A.GetLength(1); k++)
				{
					multiply = new CodeBinaryOperatorExpression();
					multiply.Operator = CodeBinaryOperatorType.Multiply;
					multiply.Right = new CodeSnippetExpression(string.Format("A[i,{0}]",k));
					multiply.Left = new CodeSnippetExpression(string.Format("B[{0},j]",k));
					if(k < A.GetLength(1) - 1)
					{
						plus = new CodeBinaryOperatorExpression();
						plus.Operator = CodeBinaryOperatorType.Add;
						lastplus.Right = plus;
						plus.Left = multiply;
					}
					else
					{
						lastplus.Right = multiply;
					}
					lastplus = plus;
				}
				multiplyHelperMethod.Statements.Add(returnStatement);
			}

			matrixClass.Members.Add(multiplyHelperMethod);

			// Set up the Multiply function
			CodeMemberMethod multiplyMethod = new CodeMemberMethod();
			multiplyMethod.Name = "Multiply";
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(double [,]), "A"));
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(double [,]), "B"));
			multiplyMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(double [,]), "C"));
			// workaround for bug below...
			multiplyMethod.ReturnType = new CodeTypeReference("public void");
			// BUG: This doesn't generate "public", it just leaves
			// the attribute off of the member...
			multiplyMethod.Attributes |= MemberAttributes.Public;

			// Loop over the rows of A
			for(int i = 0; i < A.GetLength(0); i++)
			{
				// Loop over the columns of B
				for(int j = 0; j < B.GetLength(1); j++)
				{
					CodeAssignStatement assignment = new CodeAssignStatement();
					assignment.Left = new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("C"), 
						                                             new CodePrimitiveExpression(i), new CodePrimitiveExpression(j));
					assignment.Right = new CodeMethodInvokeExpression(null,
                                                                      "_Multiply",
						                                              new CodeVariableReferenceExpression("A"),
						                                              new CodeVariableReferenceExpression("B"),
						                                              new CodePrimitiveExpression(i), new CodePrimitiveExpression(j));
					multiplyMethod.Statements.Add(assignment);
				}
			}

			matrixClass.Members.Add(multiplyMethod);
			cg.GenerateCodeFromType(matrixClass, t, op);

			t.Close();
			s.Close();

			string target = "mml.dll";
			// Now use the CodeProvider interface
			CSharpCodeProvider cscp = new CSharpCodeProvider();
			CompilerParameters param = new CompilerParameters(null, null, true);
			param.GenerateInMemory = true;
			param.IncludeDebugInformation = false;
			param.CompilerOptions = "/optimize+";
			param.ReferencedAssemblies.Add("MatrixMultiplication.exe");
			ICodeCompiler cc = cscp.CreateCompiler();
			CompilerResults cr = cc.CompileAssemblyFromFile(param, filename + ".cs");
			System.Collections.Specialized.StringCollection output = cr.Output;
			foreach(string outputSting in output)
			{
				Console.WriteLine(outputSting);
			}

			if(cr.Errors.Count != 0)
			{
				CompilerErrorCollection es = cr.Errors;
				foreach(System.CodeDom.Compiler.CompilerError e in es)
				{
					Console.WriteLine(e.ToString());
				}
			}
			else
			{
				matrixMultiplyInterface = (IMatrixMultiply)cr.CompiledAssembly.CreateInstance(className);
			}
		}
		public override double [,] Multiply()
		{
			if (matrixMultiplyInterface == null)
				throw new ArgumentOutOfRangeException("Could not compile fixed matrix code.");
			try
			{
				matrixMultiplyInterface.Multiply(A, B, C);
			}
			catch(Exception e)
			{
				Console.WriteLine(e.ToString());
				throw;
			}
			return C;
		}
	}
}

